using Polymake
using DynamicPolynomials
using LinearAlgebra
using HomotopyContinuation
using IterTools
include("PolytopeUtils.jl")
using HCubature

# This computes the integral using Monte Carlo sampling from the tropical distribution.
function integrate_tropical(F,numers,denoms,x,μ,ν,nsamples; verbatim = false)
    #---------------------------------
    if verbatim println("1. Sum the Newton polytopes") end
    #---------------------------------

    P = [newtonPolytope(numer,x) for numer ∈ numers]
    Q = [newtonPolytope(denom,x) for denom ∈ denoms]
    MS, vtcs = sumAll([P;Q])

    #---------------------------------
    if verbatim println("----------------------------------------------") end
    if verbatim println("2. Compute the normal fan of the sum") end
    #---------------------------------

    Σ = Polymake.fan.normal_fan(MS)
    if !Σ.SIMPLICIAL
        A, b = facetRepresentation(MS)
        b = b + rand(1:99,length(b)).//10000
        MS = @pm polytope.Polytope(FACETS = [b A])
        Σ = Polymake.fan.normal_fan(MS)
    end
    
    rays = Σ.RAYS
    maxcones = Σ.MAXIMAL_CONES
    maxcones = convert(Array{Int64,2},maxcones)

    #---------------------------------
    if verbatim println("Σ has $(size(rays,1)) rays and $(size(maxcones,1)) maximal cones") end
    if verbatim println("----------------------------------------------") end
    if verbatim println("3. For each maximal cone, compute the winning exponent") end
    #---------------------------------

    rep_rays = [sum([rays[i,:] for i ∈ findall(ℓ->ℓ==1,maxcones[k,:])]) for k = 1:size(maxcones,1)]
    EM_numers = [hcat(exponents.(terms(numer+0*sum(x)))...) for numer ∈ numers]
    EM_denoms = [hcat(exponents.(terms(denom+0*sum(x)))...) for denom ∈ denoms]
    ws = [sum([μ[ℓ]*EM_numers[ℓ][:,argmax((-rep_rays[i]'*EM_numers[ℓ])[:])] for ℓ = 1:length(numers)]) - sum([ν[ℓ]*EM_denoms[ℓ][:,argmax((-rep_rays[i]'*EM_denoms[ℓ])[:])] for ℓ = 1:length(denoms)]) for i = 1:length(rep_rays) ]


    #---------------------------------
    if verbatim println(ws) end
    if verbatim println("----------------------------------------------") end
    if verbatim println("4. Compute the sector integral for all $(size(maxcones,1)) maximal cones") end
    #---------------------------------

    n, k = size(F)
    Is = collect(subsets(1:k,n))
    Int_approx = 0
    S = System([numers;denoms] + 0.0*[numers;denoms])

    prefactor = fill(zero(Rational{Int64}),size(maxcones,1))
    for k = 1:size(maxcones,1)
        wσ = ws[k] # winning exponent
        rayinds = findall(ℓ->ℓ==1, maxcones[k,:]) # ray indices for this cone
        U = convert(Array{Rational{Int64},2},transpose(rays[rayinds,:])) # flip the rays (max-convention) and collect in a matrix
        prefactor[k] = sum([abs(det(F[:,I]*U[I,:])) for I ∈ Is])/prod([dot(U[:,i],wσ) for i = 1:n])
    end
    Itr = sum(prefactor)
    if verbatim println("     <<<<<<< Itr = $(Itr) >>>>>>") end

    for k = 1:size(maxcones,1)
        wσ = ws[k] # winning exponent
        rayinds = findall(ℓ->ℓ==1, maxcones[k,:]) # ray indices for this cone
        U = convert(Array{Rational{Int64},2},transpose(rays[rayinds,:])) # flip the rays (max-convention) and collect in a matrix
        Ĩ = 0
        nsamples_σ = convert(Int64,round(prefactor[k]/Itr*nsamples))
        if verbatim println("    I_σ^tr = $(prefactor[k])") end
        if verbatim println("    Number of samples for this cone: $(nsamples_σ)") end
        for i = 1:nsamples_σ
            q = rand(n)
            λ = [1//dot(U[:,i],wσ)*log(q[i]) for i = 1:length(q)]
            y = U*λ
            xx = exp.(y)
            vals = [convert.(Float64,subs(numers, x=>xx)); convert.(Float64,subs(denoms, x=>xx))]
            logff = sum([μ[i]*log(vals[i]) for i = 1:length(numers)]) - sum([ν[i-length(numers)]*log(vals[i]) for i = length(numers)+1:length(numers)+length(denoms)])
            #println(logff)
            ff = exp(logff)*prod(xx.^(-wσ))
            Ĩ += ff
        end
        this_int = prefactor[k]*Ĩ/nsamples_σ
        Int_approx += this_int

        #---------------------------------
        if verbatim println("    Computed the value for sector $(k): $(this_int)") end
        #---------------------------------
    end

    #---------------------------------
    if verbatim println("----------------------------------------------") end
    if verbatim println("5. The integral is $(Int_approx)") end
    #---------------------------------

    Int_approx
end

function sample_tropical(F,numers,denoms,x,μ,ν,nsamples; verbatim = false)
    #---------------------------------
    if verbatim println("1. Sum the Newton polytopes") end
    #---------------------------------

    P = [newtonPolytope(numer,x) for numer ∈ numers]
    Q = [newtonPolytope(denom,x) for denom ∈ denoms]
    MS, vtcs = sumAll([P;Q])

    #---------------------------------
    if verbatim println("----------------------------------------------") end
    if verbatim println("2. Compute the normal fan of the sum") end
    #---------------------------------

    Σ = Polymake.fan.normal_fan(MS)
    if !Σ.SIMPLICIAL
        A, b = facetRepresentation(MS)
        b = b + rand(1:99,length(b)).//10000
        MS = @pm polytope.Polytope(FACETS = [b A])
        Σ = Polymake.fan.normal_fan(MS)
    end
    
    rays = Σ.RAYS
    maxcones = Σ.MAXIMAL_CONES
    maxcones = convert(Array{Int64,2},maxcones)

    #---------------------------------
    if verbatim println("Σ has $(size(rays,1)) rays and $(size(maxcones,1)) maximal cones") end
    if verbatim println("----------------------------------------------") end
    if verbatim println("3. For each maximal cone, compute the winning exponent") end
    #---------------------------------

    rep_rays = [sum([rays[i,:] for i ∈ findall(ℓ->ℓ==1,maxcones[k,:])]) for k = 1:size(maxcones,1)]
    EM_numers = [hcat(exponents.(terms(numer+0*sum(x)))...) for numer ∈ numers]
    EM_denoms = [hcat(exponents.(terms(denom+0*sum(x)))...) for denom ∈ denoms]
    ws = [sum([μ[ℓ]*EM_numers[ℓ][:,argmax((-rep_rays[i]'*EM_numers[ℓ])[:])] for ℓ = 1:length(numers)]) - sum([ν[ℓ]*EM_denoms[ℓ][:,argmax((-rep_rays[i]'*EM_denoms[ℓ])[:])] for ℓ = 1:length(denoms)]) for i = 1:length(rep_rays) ]


    #---------------------------------
    if verbatim println(ws) end
    if verbatim println("----------------------------------------------") end
    if verbatim println("4. Compute the sector integral for all $(size(maxcones,1)) maximal cones") end
    #---------------------------------

    n, k = size(F)
    Is = collect(subsets(1:k,n))
    Int_approx = 0

    prefactor = fill(zero(Rational{Int64}),size(maxcones,1))
    for k = 1:size(maxcones,1)
        wσ = ws[k] # winning exponent
        rayinds = findall(ℓ->ℓ==1, maxcones[k,:]) # ray indices for this cone
        U = convert(Array{Rational{Int64},2},transpose(rays[rayinds,:])) # flip the rays (max-convention) and collect in a matrix
        prefactor[k] = sum([abs(det(F[:,I]*U[I,:])) for I ∈ Is])/prod([dot(U[:,i],wσ) for i = 1:n])
    end
    Itr = sum(prefactor)
    if verbatim println("     <<<<<<< Itr = $(Itr) >>>>>>") end
    
    samples = []
    densities = []
    for k = 1:size(maxcones,1)
        wσ = ws[k] # winning exponent
        rayinds = findall(ℓ->ℓ==1, maxcones[k,:]) # ray indices for this cone
        U = convert(Array{Rational{Int64},2},transpose(rays[rayinds,:])) # flip the rays (max-convention) and collect in a matrix
        nsamples_σ = convert(Int64,round(prefactor[k]/Itr*nsamples))
        if verbatim println("    I_σ^tr = $(prefactor[k])") end
        if verbatim println("    Number of samples for this cone: $(nsamples_σ)") end
        for i = 1:nsamples_σ
            q = rand(n)
            λ = [1//dot(U[:,i],wσ)*log(q[i]) for i = 1:length(q)]
            y = U*λ
            xx = exp.(y)
            samples = push!(samples,xx)
            densities = push!(densities,prod(xx.^(-wσ)))
        end
    end

    samples, densities
end

function sample_rejection(F,numers,denoms,x,μ,ν,nsamples; verbatim = false)
    
    #---------------------------------
    if verbatim println("0. Compute the bounds M1, M2 and the integral I") end
    #---------------------------------
    M1 = getM1(numers, denoms, μ, ν)
    M2 = getM2(numers, denoms, μ, ν)
    #I = integrate_tropical(F,numers,denoms,x,μ,ν,nsamples; verbatim = false)
    if verbatim println("     M1 = $(M1), M2 = $(M2)") end
    
    #---------------------------------
    if verbatim println("1. Sum the Newton polytopes") end
    #---------------------------------

    P = [newtonPolytope(numer,x) for numer ∈ numers]
    Q = [newtonPolytope(denom,x) for denom ∈ denoms]
    MS, vtcs = sumAll([P;Q])

    #---------------------------------
    if verbatim println("----------------------------------------------") end
    if verbatim println("2. Compute the normal fan of the sum") end
    #---------------------------------

    Σ = Polymake.fan.normal_fan(MS)
    if !Σ.SIMPLICIAL
        A, b = facetRepresentation(MS)
        b = b + rand(1:99,length(b)).//10000
        MS = @pm polytope.Polytope(FACETS = [b A])
        Σ = Polymake.fan.normal_fan(MS)
    end
    
    rays = Σ.RAYS
    maxcones = Σ.MAXIMAL_CONES
    maxcones = convert(Array{Int64,2},maxcones)

    #---------------------------------
    if verbatim println("Σ has $(size(rays,1)) rays and $(size(maxcones,1)) maximal cones") end
    if verbatim println("----------------------------------------------") end
    if verbatim println("3. For each maximal cone, compute the winning exponent") end
    #---------------------------------

    rep_rays = [sum([rays[i,:] for i ∈ findall(ℓ->ℓ==1,maxcones[k,:])]) for k = 1:size(maxcones,1)]
    EM_numers = [hcat(exponents.(terms(numer+0*sum(x)))...) for numer ∈ numers]
    EM_denoms = [hcat(exponents.(terms(denom+0*sum(x)))...) for denom ∈ denoms]
    ws = [sum([μ[ℓ]*EM_numers[ℓ][:,argmax((-rep_rays[i]'*EM_numers[ℓ])[:])] for ℓ = 1:length(numers)]) - sum([ν[ℓ]*EM_denoms[ℓ][:,argmax((-rep_rays[i]'*EM_denoms[ℓ])[:])] for ℓ = 1:length(denoms)]) for i = 1:length(rep_rays) ]


    #---------------------------------
    if verbatim println(ws) end
    if verbatim println("----------------------------------------------") end
    if verbatim println("4. Compute the sector integral for all $(size(maxcones,1)) maximal cones") end
    #---------------------------------

    n, k = size(F)
    Is = collect(subsets(1:k,n))
    Int_approx = 0

    prefactor = fill(zero(Rational{Int64}),size(maxcones,1))
    for k = 1:size(maxcones,1)
        wσ = ws[k] # winning exponent
        rayinds = findall(ℓ->ℓ==1, maxcones[k,:]) # ray indices for this cone
        U = convert(Array{Rational{Int64},2},transpose(rays[rayinds,:])) # flip the rays (max-convention) and collect in a matrix
        prefactor[k] = sum([abs(det(F[:,I]*U[I,:])) for I ∈ Is])/prod([dot(U[:,i],wσ) for i = 1:n])
    end
    Itr = sum(prefactor)
    if verbatim println("     <<<<<<< Itr = $(Itr) >>>>>>") end
    
    #C = Itr/I*M2
    samples = []
    densities = []
    for k = 1:size(maxcones,1)
        wσ = ws[k] # winning exponent
        rayinds = findall(ℓ->ℓ==1, maxcones[k,:]) # ray indices for this cone
        U = convert(Array{Rational{Int64},2},transpose(rays[rayinds,:])) # flip the rays (max-convention) and collect in a matrix
        nsamples_σ = convert(Int64,round(prefactor[k]/Itr*nsamples))
        if verbatim println("    I_σ^tr = $(prefactor[k])") end
        if verbatim println("    Number of samples for this cone: $(nsamples_σ)") end
        for i = 1:nsamples_σ
            q = rand(n)
            λ = [1//dot(U[:,i],wσ)*log(q[i]) for i = 1:length(q)]
            y = U*λ
            xx = exp.(y)
            vals = [convert.(Float64,subs(numers, x=>xx)); convert.(Float64,subs(denoms, x=>xx))]
            logff = sum([μ[i]*log(vals[i]) for i = 1:length(numers)]) - sum([ν[i-length(numers)]*log(vals[i]) for i = length(numers)+1:length(numers)+length(denoms)])
            ff = exp(logff)*prod(xx.^(-wσ))
            xi = rand()*M2
            if xi < ff 
                #println("$(xi) < $((Itr/I)*ff)")
                samples = push!(samples,xx)
                densities = push!(densities,prod(xx.^(-wσ)))
            end
        end
    end

    samples, densities
end



# This function uses numerical cubature for evaluating the sector integrals up to precision inttol.
function integrate_cubature(F,numers,denoms,x,μ,ν; verbatim = false, inttol = 1e-3)
    #---------------------------------
    if verbatim println("1. Sum the Newton polytopes") end
    #---------------------------------

    P = [newtonPolytope(numer,x) for numer ∈ numers]
    Q = [newtonPolytope(denom,x) for denom ∈ denoms]
 
    MS, vtcs = sumAll([P;Q])

    #---------------------------------
    if verbatim println("----------------------------------------------") end
    if verbatim println("2. Compute the normal fan of the sum") end
    #---------------------------------

    Σ = Polymake.fan.normal_fan(MS)
    if !Σ.SIMPLICIAL
        A, b = facetRepresentation(MS)
        b = b + rand(1:99,length(b)).//10000
        MS = @pm polytope.Polytope(FACETS = [b A])
        Σ = Polymake.fan.normal_fan(MS)
    end
    #println(Σ.SIMPLICIAL)
    rays = Σ.RAYS
    maxcones = Σ.MAXIMAL_CONES
    maxcones = convert(Array{Int64,2},maxcones)

    #---------------------------------
    if verbatim println("Σ has $(length(rays)) rays and $(size(maxcones,1)) maximal cones") end
    if verbatim println("----------------------------------------------") end
    if verbatim println("3. For each maximal cone, compute the winning exponent") end
    #---------------------------------

    rep_rays = [sum([rays[i,:] for i ∈ findall(ℓ->ℓ==1,maxcones[k,:])]) for k = 1:size(maxcones,1)]
    EM_numers = [hcat(exponents.(terms(numer+0*sum(x)))...) for numer ∈ numers]
    EM_denoms = [hcat(exponents.(terms(denom+0*sum(x)))...) for denom ∈ denoms]
    ws = [sum([μ[ℓ]*EM_numers[ℓ][:,argmax((-rep_rays[i]'*EM_numers[ℓ])[:])] for ℓ = 1:length(numers)]) - sum([ν[ℓ]*EM_denoms[ℓ][:,argmax((-rep_rays[i]'*EM_denoms[ℓ])[:])] for ℓ = 1:length(denoms)]) for i = 1:length(rep_rays) ]


    #---------------------------------
    if verbatim println(ws) end
    if verbatim println("----------------------------------------------") end
    if verbatim println("4. Compute the sector integral for all $(size(maxcones,1)) maximal cones") end
    #---------------------------------

    n, k = size(F)
    Is = collect(subsets(1:k,n))
    Int_approx = 0
    S = System([numers;denoms])
    #println(S)

    function ϕ(n,U,wσ,S,q)
        λ = [1//dot(U[:,i],wσ)*log(q[i]) for i = 1:length(q)]
        y = U*λ
        xx = exp.(y)
        #vals = S(xx)
        vals = [convert.(Float64,subs(numers, x=>xx)); convert.(Float64,subs(denoms, x=>xx))]
        logff = sum([μ[i]*log(vals[i]) for i = 1:length(numers)]) - sum([ν[i-length(numers)]*log(vals[i]) for i = length(numers)+1:length(numers)+length(denoms)])
        ff = exp(logff)*prod(xx.^(-wσ))
    end

    for k = 1:size(maxcones,1)
        wσ = ws[k] # winning exponent
        rayinds = findall(ℓ->ℓ==1, maxcones[k,:]) # ray indices for this cone
        U = convert(Array{Rational{Int64},2},transpose(rays[rayinds,:])) # flip the rays (max-convention) and collect in a matrix
        prefactor = abs(det(F*U))/prod([dot(U[:,i],wσ) for i = 1:n])
        #println("tropical value: $(prefactor)")
        func = q -> ϕ(n,U,wσ,S,q);
        (this_int, err) = HCubature.hcubature(func, zeros(n), ones(n); rtol = inttol)
        Int_approx += prefactor*this_int

        #---------------------------------
        if verbatim println("    Computed the value for sector $(k): $(this_int*prefactor)"); end #println("    Error: $(err)"); println("    Combinatorial factor: $(prefactor)"); end
        #---------------------------------
    end

    #---------------------------------
    if verbatim println("----------------------------------------------") end
    if verbatim println("5. The integral is $(Int_approx)") end
    #---------------------------------

    Int_approx
end

function getM1(numers, denoms, mu, nu)
    M1_num = BigFloat(1.0)
    M1_den = BigFloat(1.0)
    for i = 1:length(numers)
        M1_num = M1_num*(minimum(coefficients(numers[i])))^mu[i]
    end
    for j = 1:length(denoms)
        M1_den = M1_den*(sum(coefficients(denoms[j])))^nu[j]
    end
    M1_num/M1_den
end

function getM2(numers, denoms, mu, nu)
    M2_num = BigFloat(1.0)
    M2_den = BigFloat(1.0)
    for i = 1:length(numers)
        M2_num = M2_num*(sum(coefficients(numers[i])))^mu[i]
    end
    for j = 1:length(denoms)
        M2_den = M2_den*(minimum(coefficients(denoms[j])))^nu[j]
    end
    M2_num/M2_den
end


